/*
 *        spString.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spString.h>

#if defined(_WIN32)
#include <windows.h>
#if !defined(__CYGWIN32__)
#include <mbctype.h>
#include <mbstring.h>
#define strchr(string, c) ((char *)_mbschr(((unsigned char *)string), (unsigned int)c))
#define strrchr(string, c) ((char *)_mbsrchr(((unsigned char *)string), (unsigned int)c))
#endif
#endif

#if defined(IPHONE) || defined(MACOSX64)
#define SP_MAC_USE_CFSTR_CONVERTER
#include <CoreFoundation/CoreFoundation.h>
#elif defined(MACOSX)
#include <CoreServices/CoreServices.h>
#elif defined(MACOS)
#include <Gestalt.h>
#include <Script.h>
#include <TextEncodingConverter.h>
#include <MacErrors.h>
#endif

static char *locale_alias[][2] = {
    /* based on locale.alias of X window system */
    {"POSIX", "C"},
    {"POSIX-UTF2", "C"},
    {"C_C.C", "C"},
    {"C.en", "C"},
    {"C.iso88591", "en_US.ISO8859-1"},
    {"C.iso885915", "en_US.ISO8859-15"},
    {"Cextend", "en_US.ISO8859-1"},
    {"Cextend.en", "en_US.ISO8859-1"},
    {"English_United-States.437", "C"},
    {"ar", "ar_AA.ISO8859-6"},
    {"ar_AA", "ar_AA.ISO8859-6"},
    {"ar_SA.iso88596", "ar_SA.ISO8859-6"},
    {"bg", "bg_BG.ISO8859-5"},
    {"bg_BG", "bg_BG.ISO8859-5"},
    {"bg_BG.iso88595", "bg_BG.ISO8859-5"},
    {"cs", "cs_CZ.ISO8859-2"},
    {"cs_CS", "cs_CZ.ISO8859-2"},
    {"cs_CS.ISO8859-2", "cs_CZ.ISO8859-2"},
    {"cs_CZ", "cs_CZ.ISO8859-2"},
    {"cs_CZ.iso88592", "cs_CZ.ISO8859-2"},
    {"cz", "cz_CZ.ISO8859-2"},
    {"cz_CZ", "cz_CZ.ISO8859-2"},
    {"da", "da_DK.ISO8859-1"},
    {"da_DK", "da_DK.ISO8859-1"},
    {"da_DK.88591", "da_DK.ISO8859-1"},
    {"da_DK.885915", "da_DK.ISO8859-15"},
    {"da_DK.88591.en", "da_DK.ISO8859-1"},
    {"da_DK.885915.en", "da_DK.ISO8859-15"},
    {"da_DK.iso88591", "da_DK.ISO8859-1"},
    {"da_DK.iso885915", "da_DK.ISO8859-15"},
    {"da_DK.ISO_8859-1", "da_DK.ISO8859-1"},
    {"da_DK.ISO_8859-15", "da_DK.ISO8859-15"},
    {"de", "de_DE.ISO8859-1"},
    {"de_AT", "de_AT.ISO8859-1"},
    {"de_AT.ISO_8859-1", "de_AT.ISO8859-1"},
    {"de_AT.ISO_8859-15", "de_AT.ISO8859-15"},
    {"de_CH", "de_CH.ISO8859-1"},
    {"de_CH.ISO_8859-1", "de_CH.ISO8859-1"},
    {"de_CH.ISO_8859-15", "de_CH.ISO8859-15"},
    {"de_DE", "de_DE.ISO8859-1"},
    {"de_DE.88591", "de_DE.ISO8859-1"},
    {"de_DE.885915", "de_DE.ISO8859-15"},
    {"de_DE.88591.en", "de_DE.ISO8859-1"},
    {"de_DE.885915.en", "de_DE.ISO8859-15"},
    {"de_DE.iso88591", "de_DE.ISO8859-1"},
    {"de_DE.iso885915", "de_DE.ISO8859-15"},
    {"de_DE.ISO_8859-1", "de_DE.ISO8859-1"},
    {"de_DE.ISO_8859-15", "de_DE.ISO8859-15"},
    {"GER_DE.8859", "de_DE.ISO8859-1"},
    {"GER_DE.8859.in", "de_DE.ISO8859-1"},
    {"ee", "ee_EE.ISO8859-4"},
    {"el", "el_GR.ISO8859-7"},
    {"el_GR", "el_GR.ISO8859-7"},
    {"el_GR.iso88597", "el_GR.ISO8859-7"},
    {"en", "en_US.ISO8859-1"},
    {"en_AU", "en_AU.ISO8859-1"},
    {"en_AU.ISO_8859-1", "en_AU.ISO8859-1"},
    {"en_AU.ISO_8859-15", "en_AU.ISO8859-15"},
    {"en_CA", "en_CA.ISO8859-1"},
    {"en_CA.ISO_8859-1", "en_CA.ISO8859-1"},
    {"en_CA.ISO_8859-15", "en_CA.ISO8859-15"},
    {"en_GB", "en_GB.ISO8859-1"},
    {"en_GB.88591", "en_GB.ISO8859-1"},
    {"en_GB.885915", "en_GB.ISO8859-15"},
    {"en_GB.88591.en", "en_GB.ISO8859-1"},
    {"en_GB.885915.en", "en_GB.ISO8859-15"},
    {"en_GB.iso88591", "en_GB.ISO8859-1"},
    {"en_GB.iso885915", "en_GB.ISO8859-15"},
    {"en_GB.ISO_8859-1", "en_GB.ISO8859-1"},
    {"en_GB.ISO_8859-15", "en_GB.ISO8859-15"},
    {"en_UK", "en_GB.ISO8859-1"},
    {"ENG_GB.8859", "en_GB.ISO8859-1"},
    {"ENG_GB.8859.in", "en_GB.ISO8859-1"},
    {"en_IE", "en_IE.ISO8859-1"},
    {"en_NZ", "en_NZ.ISO8859-1"},
    {"en_US", "en_US.ISO8859-1"},
    {"en_US.88591", "en_US.ISO8859-1"},
    {"en_US.885915", "en_US.ISO8859-15"},
    {"en_US.88591.en", "en_US.ISO8859-1"},
    {"en_US.885915.en", "en_US.ISO8859-15"},
    {"en_US.iso88591", "en_US.ISO8859-1"},
    {"en_US.iso885915", "en_US.ISO8859-15"},
    {"en_US.ISO_8859-1", "en_US.ISO8859-1"},
    {"en_US.ISO_8859-15", "en_US.ISO8859-15"},
    {"en_US.UTF-8", "en_US.utf"},
    {"es", "es_ES.ISO8859-1"},
    {"es_AR", "es_AR.ISO8859-1"},
    {"es_BO", "es_BO.ISO8859-1"},
    {"es_CL", "es_CL.ISO8859-1"},
    {"es_CO", "es_CO.ISO8859-1"},
    {"es_CR", "es_CR.ISO8859-1"},
    {"es_EC", "es_EC.ISO8859-1"},
    {"es_ES", "es_ES.ISO8859-1"},
    {"es_ES.88591", "es_ES.ISO8859-1"},
    {"es_ES.885915", "es_ES.ISO8859-15"},
    {"es_ES.88591.en", "es_ES.ISO8859-1"},
    {"es_ES.885915.en", "es_ES.ISO8859-15"},
    {"es_ES.iso88591", "es_ES.ISO8859-1"},
    {"es_ES.iso885915", "es_ES.ISO8859-15"},
    {"es_ES.ISO_8859-1", "es_ES.ISO8859-1"},
    {"es_ES.ISO_8859-15", "es_ES.ISO8859-15"},
    {"es_GT", "es_GT.ISO8859-1"},
    {"es_MX", "es_MX.ISO8859-1"},
    {"es_NI", "es_NI.ISO8859-1"},
    {"es_PA", "es_PA.ISO8859-1"},
    {"es_PE", "es_PE.ISO8859-1"},
    {"es_PY", "es_PY.ISO8859-1"},
    {"es_SV", "es_SV.ISO8859-1"},
    {"es_UY", "es_UY.ISO8859-1"},
    {"es_VE", "es_VE.ISO8859-1"},
    {"et", "et_EE.ISO8859-4"},
    {"et_EE", "et_EE.ISO8859-4"},
    {"fi", "fi_FI.ISO8859-1"},
    {"fi_FI", "fi_FI.ISO8859-1"},
    {"fi_FI.88591", "fi_FI.ISO8859-1"},
    {"fi_FI.885915", "fi_FI.ISO8859-15"},
    {"fi_FI.88591.en", "fi_FI.ISO8859-1"},
    {"fi_FI.885915.en", "fi_FI.ISO8859-15"},
    {"fi_FI.iso88591", "fi_FI.ISO8859-1"},
    {"fi_FI.iso885915", "fi_FI.ISO8859-15"},
    {"fi_FI.ISO_8859-1", "fi_FI.ISO8859-1"},
    {"fi_FI.ISO_8859-15", "fi_FI.ISO8859-15"},
    {"fr", "fr_FR.ISO8859-1"},
    {"fr_BE", "fr_BE.ISO8859-1"},
    {"fr_BE.88591", "fr_BE.ISO8859-1"},
    {"fr_BE.885915", "fr_BE.ISO8859-15"},
    {"fr_BE.88591.en", "fr_BE.ISO8859-1"},
    {"fr_BE.885915.en", "fr_BE.ISO8859-15"},
    {"fr_BE.ISO_8859-1", "fr_BE.ISO8859-1"},
    {"fr_BE.ISO_8859-15", "fr_BE.ISO8859-15"},
    {"fr_CA", "fr_CA.ISO8859-1"},
    {"fr_CA.88591", "fr_CA.ISO8859-1"},
    {"fr_CA.885915", "fr_CA.ISO8859-15"},
    {"fr_CA.88591.en", "fr_CA.ISO8859-1"},
    {"fr_CA.885915.en", "fr_CA.ISO8859-15"},
    {"fr_CA.iso88591", "fr_CA.ISO8859-1"},
    {"fr_CA.iso885915", "fr_CA.ISO8859-15"},
    {"fr_CA.ISO_8859-1", "fr_CA.ISO8859-1"},
    {"fr_CA.ISO_8859-15", "fr_CA.ISO8859-15"},
    {"fr_CH", "fr_CH.ISO8859-1"},
    {"fr_CH.88591", "fr_CH.ISO8859-1"},
    {"fr_CH.885915", "fr_CH.ISO8859-15"},
    {"fr_CH.88591.en", "fr_CH.ISO8859-1"},
    {"fr_CH.885915.en", "fr_CH.ISO8859-15"},
    {"fr_CH.ISO_8859-1", "fr_CH.ISO8859-1"},
    {"fr_CH.ISO_8859-15", "fr_CH.ISO8859-15"},
    {"fr_FR", "fr_FR.ISO8859-1"},
    {"fr_FR.88591", "fr_FR.ISO8859-1"},
    {"fr_FR.885915", "fr_FR.ISO8859-15"},
    {"fr_FR.88591.en", "fr_FR.ISO8859-1"},
    {"fr_FR.885915.en", "fr_FR.ISO8859-15"},
    {"fr_FR.iso88591", "fr_FR.ISO8859-1"},
    {"fr_FR.iso885915", "fr_FR.ISO8859-15"},
    {"fr_FR.ISO_8859-1", "fr_FR.ISO8859-1"},
    {"fr_FR.ISO_8859-15", "fr_FR.ISO8859-15"},
    {"FRE_FR.8859", "fr_FR.ISO8859-1"},
    {"FRE_FR.8859.in", "fr_FR.ISO8859-1"},
    {"hr", "hr_HR.ISO8859-2"},
    {"hr_HR", "hr_HR.ISO8859-2"},
    {"hr_HR.iso88592", "hr_HR.ISO8859-2"},
    {"hr_HR.ISO_8859-2", "hr_HR.ISO8859-2"},
    {"hu", "hu_HU.ISO8859-2"},
    {"hu_HU", "hu_HU.ISO8859-2"},
    {"hu_HU.iso88592", "hu_HU.ISO8859-2"},
    {"is", "is_IS.ISO8859-1"},
    {"is_IS", "is_IS.ISO8859-1"},
    {"is_IS.iso88591", "is_IS.ISO8859-1"},
    {"is_IS.iso885915", "is_IS.ISO8859-15"},
    {"is_IS.ISO_8859-1", "is_IS.ISO8859-1"},
    {"is_IS.ISO_8859-15", "is_IS.ISO8859-15"},
    {"it", "it_IT.ISO8859-1"},
    {"it_CH", "it_CH.ISO8859-1"},
    {"it_CH.ISO_8859-1", "it_CH.ISO8859-1"},
    {"it_CH.ISO_8859-15", "it_CH.ISO8859-15"},
    {"it_IT", "it_IT.ISO8859-1"},
    {"it_IT.88591", "it_IT.ISO8859-1"},
    {"it_IT.885915", "it_IT.ISO8859-15"},
    {"it_IT.88591.en", "it_IT.ISO8859-1"},
    {"it_IT.885915.en", "it_IT.ISO8859-15"},
    {"it_IT.iso88591", "it_IT.ISO8859-1"},
    {"it_IT.iso885915", "it_IT.ISO8859-15"},
    {"it_IT.ISO_8859-1", "it_IT.ISO8859-1"},
    {"it_IT.ISO_8859-15", "it_IT.ISO8859-15"},
    {"iw", "iw_IL.ISO8859-8"},
    {"iw_IL", "iw_IL.ISO8859-8"},
    {"iw_IL.iso88598", "iw_IL.ISO8859-8"},
    {"ja", "ja_JP.eucJP"},
    {"ja.JIS", "ja_JP.JIS7"},
    {"ja.SJIS", "ja_JP.SJIS"},
    {"ja_JP", "ja_JP.eucJP"},
    {"ja_JP.ujis", "ja_JP.eucJP"},
    {"ja_JP.eucJP", "ja_JP.eucJP"},
    {"Jp_JP", "ja_JP.eucJP"},
    {"ja_JP.AJEC", "ja_JP.eucJP"},
    {"ja_JP.EUC", "ja_JP.eucJP"},
    {"ja_JP.ISO-2022-JP", "ja_JP.JIS7"},
    {"ja_JP.JIS", "ja_JP.JIS7"},
    {"ja_JP.jis7", "ja_JP.JIS7"},
    {"ja_JP.mscode", "ja_JP.SJIS"},
    {"ja_JP.SJIS", "ja_JP.SJIS"},
    {"ko", "ko_KR.eucKR"},
    {"ko_KR", "ko_KR.eucKR"},
    {"ko_KR.EUC", "ko_KR.eucKR"},
    {"ko_KR.euc", "ko_KR.eucKR"},
    {"lt", "lt_LT.ISO8859-4"},
    {"lv", "lv_LV.ISO8859-4"},
    {"mk", "mk_MK.ISO8859-5"},
    {"mk_MK", "mk_MK.ISO8859-5"},
    {"nl", "nl_NL.ISO8859-1"},
    {"nl_BE", "nl_BE.ISO8859-1"},
    {"nl_BE.88591", "nl_BE.ISO8859-1"},
    {"nl_BE.885915", "nl_BE.ISO8859-15"},
    {"nl_BE.88591.en", "nl_BE.ISO8859-1"},
    {"nl_BE.885915.en", "nl_BE.ISO8859-15"},
    {"nl_BE.ISO_8859-1", "nl_BE.ISO8859-1"},
    {"nl_BE.ISO_8859-15", "nl_BE.ISO8859-15"},
    {"nl_NL", "nl_NL.ISO8859-1"},
    {"nl_NL.88591", "nl_NL.ISO8859-1"},
    {"nl_NL.885915", "nl_NL.ISO8859-15"},
    {"nl_NL.88591.en", "nl_NL.ISO8859-1"},
    {"nl_NL.885915.en", "nl_NL.ISO8859-15"},
    {"nl_NL.iso88591", "nl_NL.ISO8859-1"},
    {"nl_NL.iso885915", "nl_NL.ISO8859-15"},
    {"nl_NL.ISO_8859-1", "nl_NL.ISO8859-1"},
    {"nl_NL.ISO_8859-15", "nl_NL.ISO8859-15"},
    {"no", "no_NO.ISO8859-1"},
    {"no_NO", "no_NO.ISO8859-1"},
    {"no_NO.88591", "no_NO.ISO8859-1"},
    {"no_NO.885915", "no_NO.ISO8859-15"},
    {"no_NO.88591.en", "no_NO.ISO8859-1"},
    {"no_NO.885915.en", "no_NO.ISO8859-15"},
    {"no_NO.iso88591", "no_NO.ISO8859-1"},
    {"no_NO.iso885915", "no_NO.ISO8859-15"},
    {"no_NO.ISO_8859-1", "no_NO.ISO8859-1"},
    {"no_NO.ISO_8859-15", "no_NO.ISO8859-15"},
    {"pl", "pl_PL.ISO8859-2"},
    {"pl_PL", "pl_PL.ISO8859-2"},
    {"pl_PL.iso88592", "pl_PL.ISO8859-2"},
    {"pt", "pt_PT.ISO8859-1"},
    {"pt_BR", "pt_BR.ISO8859-1"},
    {"pt_PT", "pt_PT.ISO8859-1"},
    {"pt_PT.88591", "pt_PT.ISO8859-1"},
    {"pt_PT.885915", "pt_PT.ISO8859-15"},
    {"pt_PT.88591.en", "pt_PT.ISO8859-1"},
    {"pt_PT.885915.en", "pt_PT.ISO8859-15"},
    {"pt_PT.iso88591", "pt_PT.ISO8859-1"},
    {"pt_PT.iso885915", "pt_PT.ISO8859-15"},
    {"pt_PT.ISO_8859-1", "pt_PT.ISO8859-1"},
    {"pt_PT.ISO_8859-15", "pt_PT.ISO8859-15"},
    {"ro", "ro_RO.ISO8859-2"},
    {"ro_RO", "ro_RO.ISO8859-2"},
    {"ro_RO.iso88592", "ro_RO.ISO8859-2"},
    {"ru", "ru_RU.ISO8859-5"},
    {"ru_RU", "ru_RU.ISO8859-5"},
    {"ru_RU.iso88595", "ru_RU.ISO8859-5"},
    {"sh", "sh_YU.ISO8859-2"},
    {"sh_HR.iso88592", "sh_HR.ISO8859-2"},
    {"sh_YU", "sh_YU.ISO8859-2"},
    {"sh_SP", "sh_YU.ISO8859-2"},
    {"sk", "sk_SK.ISO8859-2"},
    {"sk_SK", "sk_SK.ISO8859-2"},
    {"sk_sK.iso88592", "sk_SK.ISO8859-2"},
    {"sl", "sl_CS.ISO8859-2"},
    {"sl_CS", "sl_CS.ISO8859-2"},
    {"sl_SI", "sl_SI.ISO8859-2"},
    {"sl_SI.iso88592", "sl_SI.ISO8859-2"},
    {"sp", "sp_YU.ISO8859-5"},
    {"sp_YU", "sp_YU.ISO8859-5"},
    {"sr_SP", "sr_SP.ISO8859-2"},
    {"sv", "sv_SE.ISO8859-1"},
    {"sv_SE", "sv_SE.ISO8859-1"},
    {"sv_SE.88591", "sv_SE.ISO8859-1"},
    {"sv_SE.885915", "sv_SE.ISO8859-15"},
    {"sv_SE.88591.en", "sv_SE.ISO8859-1"},
    {"sv_SE.885915.en", "sv_SE.ISO8859-15"},
    {"sv_SE.iso88591", "sv_SE.ISO8859-1"},
    {"sv_SE.iso885915", "sv_SE.ISO8859-15"},
    {"sv_SE.ISO_8859-1", "sv_SE.ISO8859-1"},
    {"sv_SE.ISO_8859-15", "sv_SE.ISO8859-15"},
    {"th_TH", "th_TH.TACTIS"},
    {"th_TH.tis620", "th_TH.TACTIS"},
    {"tr", "tr_TR.ISO8859-9"},
    {"tr_TR", "tr_TR.ISO8859-9"},
    {"tr_TR.iso88599", "tr_TR.ISO8859-9"},
    {"zh_CN", "zh_CN.GBK"},
    {"zh_CN.EUC", "zh_CN.eucCN"},
    {"zh_TW", "zh_TW.eucTW"},
    {"zh_TW.EUC", "zh_TW.eucTW"},
    {"zh_CN.big5", "zh_TW.eucTW"},
    {"english_uk.8859", "en_GB.ISO8859-1"},
    {"english_us.8859", "en_US.ISO8859-1"},
    {"english_us.ascii", "en_US.ISO8859-1"},
    {"french_france.8859", "fr_FR.ISO8859-1"},
    {"german_germany.8859", "de_DE.ISO8859-1"},
    {"portuguese_brazil.8859", "pt_BR.ISO8859-1"},
    {"spanish_spain.8859", "es_ES.ISO8859-1"},
    {"american.iso88591", "en_US.ISO8859-1"},
    {"american.iso885915", "en_US.ISO8859-15"},
    {"arabic.iso88596", "ar_AA.ISO8859-6"},
    {"bulgarian", "bg_BG.ISO8859-5"},
    {"c-french.iso88591", "fr_CA.ISO8859-1"},
    {"c-french.iso885915", "fr_CA.ISO8859-15"},
    {"chinese-s", "zh_CN.eucCN"},
    {"chinese-t", "zh_TW.eucTW"},
    {"croatian", "hr_HR.ISO8859-2"},
    {"czech", "cs_CZ.ISO8859-2"},
    {"danish.iso88591", "da_DK.ISO8859-1"},
    {"danish.iso885915", "da_DK.ISO8859-15"},
    {"dutch.iso88591", "nl_BE.ISO8859-1"},
    {"dutch.iso885915", "nl_BE.ISO8859-15"},
    {"english.iso88591", "en_EN.ISO8859-1"},
    {"english.iso885915", "en_EN.ISO8859-15"},
    {"finnish.iso88591", "fi_FI.ISO8859-1"},
    {"finnish.iso885915", "fi_FI.ISO8859-15"},
    {"french.iso88591", "fr_CH.ISO8859-1"},
    {"french.iso885915", "fr_CH.ISO8859-15"},
    {"german.iso88591", "de_CH.ISO8859-1"},
    {"german.iso885915", "de_CH.ISO8859-15"},
    {"greek.iso88597", "el_GR.ISO8859-7"},
    {"hebrew.iso88598", "iw_IL.ISO8859-8"},
    {"hungarian", "hu_HU.ISO8859-2"},
    {"icelandic.iso88591", "is_IS.ISO8859-1"},
    {"icelandic.iso885915", "is_IS.ISO8859-15"},
    {"italian.iso88591", "it_IT.ISO8859-1"},
    {"italian.iso885915", "it_IT.ISO8859-15"},
    {"japanese", "ja_JP.SJIS"},
    {"japanese.euc", "ja_JP.eucJP"},
    {"Japanese_Japan.932", "ja_JP.SJIS"},
    {"korean", "ko_KR.eucKR"},
    {"norwegian.iso88591", "no_NO.ISO8859-1"},
    {"norwegian.iso885915", "no_NO.ISO8859-15"},
    {"polish", "pl_PL.ISO8859-2"},
    {"portuguese.iso88591", "pt_PT.ISO8859-1"},
    {"portuguese.iso885915", "pt_PT.ISO8859-15"},
    {"rumanian", "ro_RO.ISO8859-2"},
    {"russian", "ru_RU.ISO8859-5"},
    {"serbocroatian", "sh_YU.ISO8859-2"},
    {"slovak", "sk_SK.ISO8859-2"},
    {"slovene", "sl_CS.ISO8859-2"},
    {"spanish.iso88591", "es_ES.ISO8859-1"},
    {"spanish.iso885915", "es_ES.ISO8859-15"},
    {"swedish.iso88591", "sv_SE.ISO8859-1"},
    {"swedish.iso885915", "sv_SE.ISO8859-15"},
    {"turkish.iso88599", "tr_TR.ISO8859-9"},
    {"univ.utf8", "en_US.utf"},
    {"universal.utf8@ucs4", "en_US.utf"},
    {"en_US.UTF-8", "en_US.utf"},
    {"iso_8859_1", "en_US.ISO8859-1"},
    {"iso_8859_15", "en_US.ISO8859-15"},
    {"ISO8859-1", "en_US.ISO8859-1"},
    {"ISO8859-15", "en_US.ISO8859-15"},
    {"ISO-8859-1", "en_US.ISO8859-1"},
    {"ISO-8859-15", "en_US.ISO8859-15"},
    {"japan", "ja_JP.eucJP"},
    {"Japanese-EUC", "ja_JP.eucJP"},
    {"ja_JP.UTF8", "ja_JP.utf8"},
    {"ja_JP.UTF-8", "ja_JP.utf8"},
    {"", ""},
};

static const char *getExactLang(const char *lang)
{
    int i;
    char *exact = NULL;

    if (strnone(lang)) return NULL;

    for (i = 0;; i++) {
        if (strnone(locale_alias[i][0])) {
            break;
        }

        if (streq(locale_alias[i][0], lang) || streq(locale_alias[i][1], lang)) {
            exact = locale_alias[i][1];
            break;
        }
    }

    return exact;
}

spBool spEqLanguage(const char *lang1, const char *lang2)
{
    spBool flag = SP_FALSE;

    if (lang1 == NULL && lang2 == NULL) {
        return SP_TRUE;
    } else if (lang1 == NULL || lang2 == NULL) {
        return SP_FALSE;
    }

    if (streq(lang1, lang2)) {
        flag = SP_TRUE;
    } else {
        if ((lang1 = getExactLang(lang1)) != NULL) {
            if ((lang2 = getExactLang(lang2)) != NULL) {
                if (streq(lang1, lang2)) {
                    flag = SP_TRUE;
                }
            }
        }
    }

    return flag;
}

spBool spGetLanguageId(const char *lang, char *id, int id_size)
{
    char *p;
    const char *exact;
    
    if (strnone(lang) || id == NULL) return SP_FALSE;

    if ((exact = getExactLang(lang)) != NULL) {
        if (streq(exact, "C")) {
            spStrCopy(id, id_size, "en");
        } else {
            spStrCopy(id, id_size, exact);

            if ((p = strchr(id, '_')) != NULL) {
                *p = NUL;
            }
        }
        return SP_TRUE;
    }

    return SP_FALSE;
}

spBool spGetLanguageCharset(const char *lang, char *charset, int charset_size)
{
    char *p;
    const char *exact;
    
    if (strnone(lang) || charset == NULL) return SP_FALSE;

    if ((exact = getExactLang(lang)) != NULL) {
        if (streq(exact, "C")) {
            spStrCopy(charset, charset_size, "");
            return SP_TRUE;
        } else {
            if ((p = strrchr(exact, '.')) != NULL && *(p + 1) != NUL) {
                spStrCopy(charset, charset_size, p + 1);
                return SP_TRUE;
            }
        }
    }

    return SP_FALSE;
}

spBool spIsUTFLang(const char *lang)
{
    char charset[SP_MAX_LINE];
    
    if (spGetLanguageCharset(lang, charset, sizeof(charset)) == SP_TRUE) {
        spDebug(80, "spIsUTFLang", "charset = %s\n", charset);
        if (strncaseeq(charset, "utf", 3)) {
            return SP_TRUE;
        }
    }

    return SP_FALSE;
}

spBool spIsJapaneseLang(const char *lang)
{
    if (strnone(lang)) return SP_FALSE;
    
    if (spEqLanguage(lang, "ja_JP.eucJP") == SP_TRUE
        || spEqLanguage(lang, "ja_JP.SJIS") == SP_TRUE
        || spEqLanguage(lang, "ja_JP.utf8") == SP_TRUE) {
        return SP_TRUE;
    }
    
    return SP_FALSE;
}

spBool spIsUTF8First(int c, int *continuation)
{
    if (!(c & 0x80)) {
        /* ascii */
        if (continuation != NULL) *continuation = 0;
        return SP_TRUE;
    } else {
        int i;
        unsigned char mask = 0x40;

        for (i = 0; i < 5; i++) {
            if (!(mask & c)) {
                break;
            }
            mask >>= 1;
        }

        if (i == 0) {
            return SP_FALSE;
        } else {
            if (continuation != NULL) *continuation = i;
        
            return SP_TRUE;
        }
    }
}

spBool spIsMBTailCandidate(int prev_c, int c)
{
#if defined(_WIN32) && !defined(__CYGWIN32__)
    return (_ismbblead(prev_c) && _ismbbtrail(c)) ? SP_TRUE : SP_FALSE;
#elif defined(ANDROID)
    if (spIsUTF8First(c, NULL) == SP_TRUE) {
        return SP_FALSE;
    } else {
        return SP_TRUE;
    }
#elif defined(_WIN32) || defined(MACOS)
    /* Shift JIS */
    if (((prev_c >= 0x81 && prev_c <= 0x9f) || (prev_c >= 0xe0 && prev_c <= 0xfc))
        && ((c >= 0x40 && c <= 0x7e) || (c >= 0x80 && c <= 0xfc))) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
#else
    /* EUC-JP */
    if ((prev_c >= 0x8e && prev_c <= 0xfe) && (c >= 0xa1 && c <= 0xfe)) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
#endif
}

char *spStrChr(const char *str, int c)
{
#if 1 || defined(MACOS9) || (defined(_WIN32) && defined(__CYGWIN32__))
    int i;
    int prev_c;
    char *p;

    prev_c = 0;
    p = (char *)str;
    for (i = 0;; i++) {
        if (*p == NUL) {
            p = NULL;
            break;
        } else if (spIsMBTailCandidate(prev_c, *p) == SP_TRUE) {
            prev_c = 0;
        } else {
            if ((int)*p == c) {
                break;
            }
            prev_c = *p;
        }
        ++p;
    }

    return p;
#else
    return strchr(str, c);
#endif
}

char *spStrRChr(const char *str, int c)
{
#if 1 || defined(MACOS9) || (defined(_WIN32) && defined(__CYGWIN32__))
    int i;
    int prev_c;
    char *p;

    prev_c = 0;
    p = NULL;
    for (i = 0;; i++) {
        if (str[i] == NUL) {
            break;
        } else if (spIsMBTailCandidate(prev_c, str[i]) == SP_TRUE) {
            prev_c = 0;
        } else {
            if ((int)str[i] == c) {
                p = (char *)&str[i];
            }
            prev_c = str[i];
        }
    }

    return p;
#else
    return strrchr(str, c);
#endif
}

int spUnicode16ToUTF8(const unsigned short *unicode, unsigned char *buf, int buf_size)
{
    int index;
    int buf_index;
    unsigned short u;
    unsigned char c;
    
    if (unicode == NULL) return -1;

    index = 0;
    buf_index = 0;
    u = unicode[index++];

    if (buf == NULL || buf_size <= 0) {
        while (u) {
            if (u < 0x0080) {
                buf_index += 1;
            } else if (u < 0x0800) {
                buf_index += 2;
            } else {
                buf_index += 3;
            }
            u = unicode[index++];
        }
        buf_index++;
    } else {
        while (u) {
            if (u < 0x0080) {
                c = (unsigned char)u;
                buf[buf_index++] = c;
            } else if (u < 0x0800) {
                if (buf_index + 2 >= buf_size) {
                    break;
                }
                c = (unsigned char)(u >> 6) | 0xc0;
                buf[buf_index++] = c;
                c = (unsigned char)(u & 0x3f) | 0x80;
                buf[buf_index++] = c;
            } else {
                if (buf_index + 3 >= buf_size) {
                    break;
                }
                c = (unsigned char)(u >> 12) | 0xe0;
                buf[buf_index++] = c;
                c = (unsigned char)((u >> 6) & 0x3f) | 0x80;
                buf[buf_index++] = c;
                c = (unsigned char)(u & 0x3f) | 0x80;
                buf[buf_index++] = c;
            }
        
            if (buf_index >= buf_size) {
                buf_index--;
                break;
            }
        
            u = unicode[index++];
        }
    
        buf[buf_index++] = 0x00;
    }

    return buf_index;
}

int spUTF8ToUnicode16(const unsigned char *utf8, unsigned short *buf, int buf_size)
{
    int index;
    int buf_index;
    unsigned char c;
    
    if (utf8 == NULL) return -1;
    
    index = 0;
    buf_index = 0;
    c = utf8[index++];

    if (buf == NULL || buf_size <= 0) {
        while (c) {
            if ((c & 0x80) == 0) {
                buf_index++;
            } else if ((c & 0xfc) == 0xfc) {
                buf_index++;
                index += 5;
            } else if ((c & 0xf8) == 0xf8) {
                buf_index++;
                index += 4;
            } else if ((c & 0xf0) == 0xf0) {
                buf_index++;
                index += 3;
            } else if ((c & 0xe0) == 0xe0) {
                buf_index++;
                index += 2;
            } else if ((c & 0xc0) == 0xc0) {
                buf_index++;
                index += 1;
            } else {
                spDebug(80, "spUTF8ToUnicode16", "!!!!!! c = %x\n", c);
            }
            c = utf8[index++];
        }
        buf_index++;
    } else {
        while (c) {
            if ((c & 0x80) == 0) {
                buf[buf_index++] = c;
            } else if ((c & 0xfc) == 0xfc) {
                buf[buf_index++] = ' ';
                index += 5;
            } else if ((c & 0xf8) == 0xf8) {
                buf[buf_index++] = ' ';
                index += 4;
            } else if ((c & 0xf0) == 0xf0) {
                buf[buf_index++] = ' ';
                index += 3;
            } else if ((c & 0xe0) == 0xe0) {
                buf[buf_index] = (c & 0x1f) << 12;
                
                c = utf8[index++];
                buf[buf_index] |= (c & 0x3f) << 6;
                
                c = utf8[index++];
                buf[buf_index++] |= (c & 0x3f);
            } else if ((c & 0xc0) == 0xc0) {
                buf[buf_index] = (c & 0x3f) << 6;
                
                c = utf8[index++];
                buf[buf_index++] |= (c & 0x3f);
            } else {
                spDebug(80, "spUTF8ToUnicode16", "!!!!!! c = %x\n", c);
            }
            if (buf_index >= buf_size / 2) {
                buf_index--;
                break;
            }
        
            c = utf8[index++];
        }
    
        buf[buf_index++] = 0x00;
    }
    spDebug(80, "spUTF8ToUnicode16", "buf_index = %d\n", buf_index);

    return buf_index * 2;
}

int spUnicode16Length(const unsigned short *unicode)
{
    const unsigned short *ptr;
    
    if (unicode == NULL) return 0;

    ptr = unicode;

    for (; *ptr; ++ptr);

    return (int)(ptr - unicode);
}

unsigned short *spUnicode16Copy(unsigned short *dest, int destlen, const unsigned short *src)
{
    int len;
    
    if (dest == NULL || destlen <= 0) return NULL;

    if (src == NULL || src[0] == 0 || destlen <= 1) {
        dest[0] = 0;
    } else {
        len = spUnicode16Length(src);
        if (len < destlen) {
            memcpy(dest, src, (len + 1) * sizeof(unsigned short));
        } else {
            destlen--;
            memcpy(dest, src, destlen * sizeof(unsigned short));
            dest[destlen] = 0;
        }
    }
    
    return dest;
}

unsigned short *xspUnicode16Clone(const unsigned short *unicode)
{
    int alloc_len;
    unsigned short *obuf;

    alloc_len = spUnicode16Length(unicode) + 1;
    obuf = xalloc(alloc_len, unsigned short);
    memcpy(obuf, unicode, alloc_len * sizeof(unsigned short));

    return obuf;
}

#if !defined(MACOS) && !defined(_WIN32) && !defined(ANDROID)
#include <errno.h>
#include <iconv.h>
#include <langinfo.h>
#ifndef HAVE_ICONV_H
#define HAVE_ICONV_H
#endif
#endif

struct _spConverter {
#if defined(SP_MAC_USE_CFSTR_CONVERTER)
    CFStringRef cfstr;
    CFStringEncoding ienc;
    CFStringEncoding oenc;
#elif defined(MACOS)
    TextEncoding ienc;
    TextEncoding oenc;
    TECObjectRef ec;
#elif defined(_WIN32) && !defined(HAVE_ICONV_H)
    const char *icode;
    const char *ocode;
#elif defined(HAVE_ICONV_H)
    iconv_t cd;
#else
    int dummy;
#endif
};

const char *spGetSupportedEncoding(const char *code)
{
#if defined(MACOS)
    if (strnone(code)) {
        return "";
    } else if (strcaseeq(code, "UTF-8") || strcaseeq(code, "UTF8")) {
        return "UTF-8";
    } else if (strcaseeq(code, "UTF-16") || strcaseeq(code, "UTF16")) {
        return "csUnicode";
    } else if (strcaseeq(code, "SJIS") || strcaseeq(code, "mscode")) {
        return "Shift_JIS";
    }
    return code;
#elif defined(_WIN32) && !defined(HAVE_ICONV_H)
    if (strnone(code) || strcaseeq(code, "mscode")) {
        return "ACP";
    } else if (strcaseeq(code, "SJIS") || strcaseeq(code, "SHIFT-JIS") || strcaseeq(code, "CP932")) {
        return "ACP";
    } else if (strcaseeq(code, "UTF-8") || strcaseeq(code, "UTF8")) {
        return "UTF-8";
    } else if (strcaseeq(code, "UTF-16") || strcaseeq(code, "UTF16")) {
        return "UTF-16";
    } else {
        return NULL;
    }
#elif defined(HAVE_ICONV_H)
    if (strnone(code)) code = nl_langinfo(CODESET);
#else
    {
        return "";
    }
#endif

    return code;
}

#if defined(SP_MAC_USE_CFSTR_CONVERTER)
#if defined(MACOSX)
extern CFStringEncoding sp_mac_readable_encoding;
#endif

static spBool getCFStringEncoding(const char *code, CFStringEncoding *encoding)
{
    *encoding = kCFStringEncodingInvalidId;
    
    if (strnone(code)) {
#if defined(MACOSX)
        *encoding =  sp_mac_readable_encoding;
#else
        *encoding =  /*CFStringGetSystemEncoding()*/kCFStringEncodingUTF8;
#endif
    } else {
        CFStringRef cfstr;
        
        if ((cfstr = CFStringCreateWithCString(kCFAllocatorDefault,
                                               code, kCFStringEncodingUTF8)) != NULL) {
            *encoding = CFStringConvertIANACharSetNameToEncoding(cfstr);
            CFRelease(cfstr);
        }
    }
    
    if (*encoding == kCFStringEncodingInvalidId) {
        return SP_FALSE;
    } else {
        return SP_TRUE;
    }
}
#elif defined(MACOS)
static spBool getTextEncoding(const char *code, TextEncoding *encoding)
{
    Str255 pstr;

    if (strnone(code)) {
        *encoding = GetScriptManagerVariable(smSysScript);
    } else {
        spStrCToP(code, pstr);
        if (TECGetTextEncodingFromInternetName == NULL
            || TECGetTextEncodingFromInternetName(encoding, pstr) != noErr) {
            spDebug(10, "getTextEncoding", "can't get internet name: %s\n", code);
            return SP_FALSE;
        }
    }
    
    spDebug(80, "getTextEncoding", "\"%s\": encoding = %lx\n", code, *encoding);
    
    return SP_TRUE;
}
#endif

spConverter spOpenConverter(const char *icode, const char *ocode)
{
    spConverter conv = NULL;

    if (strnone(icode) && strnone(ocode)) return NULL;

    if ((icode = spGetSupportedEncoding(icode)) == NULL) {
        spDebug(10, "spOpenConverter", "spGetSupportedEncoding(icode) failed\n");
        return NULL;
    }
    if ((ocode = spGetSupportedEncoding(ocode)) == NULL) {
        spDebug(10, "spOpenConverter", "spGetSupportedEncoding(ocode) failed\n");
        return NULL;
    }
    conv = xalloc(1, struct _spConverter);
    
#if defined(SP_MAC_USE_CFSTR_CONVERTER)
    if (getCFStringEncoding(icode, &conv->ienc) == SP_FALSE
        || getCFStringEncoding(ocode, &conv->oenc) == SP_FALSE) {
        xfree(conv);
        return NULL;
    }
#elif defined(MACOS)
    if (TECCreateConverter == NULL
        || getTextEncoding(icode, &conv->ienc) == SP_FALSE
        || getTextEncoding(ocode, &conv->oenc) == SP_FALSE
        || TECCreateConverter(&conv->ec, conv->ienc, conv->oenc) != noErr) {
        xfree(conv);
        return NULL;
    }
#elif defined(_WIN32) && !defined(HAVE_ICONV_H)
    conv->icode = icode;
    conv->ocode = ocode;
#elif defined(HAVE_ICONV_H)
    conv->cd = iconv_open(ocode, icode);
    if (conv->cd == (iconv_t)(-1)) {
        xfree(conv);
        return NULL;
    }
#else
    {
        xfree(conv);
        return NULL;
    }
#endif
    
    spDebug(80, "spOpenConverter", "done\n");
    
    return conv;
}

char *xspConvert(spConverter conv, const char *input)
{
    char *buf = NULL;

    if (conv == NULL || input == NULL) return NULL;
    
#if defined(SP_MAC_USE_CFSTR_CONVERTER)
    {
        CFStringRef cfstr;
        
        if ((cfstr = CFStringCreateWithCString(kCFAllocatorDefault, input, conv->ienc)) != NULL) {
            int buflen;
            
            buflen = CFStringGetMaximumSizeForEncoding(CFStringGetLength(cfstr), conv->oenc);
            spDebug(80, "xspConvert", "buflen = %d\n", buflen);
            
            buf = xalloc(buflen + 1, char);
            if (!CFStringGetCString(cfstr, buf, buflen + 1, conv->oenc)) {
                spDebug(10, "xspConvert", "CFStringGetCString failed\n");
                xfree(buf); buf = NULL;
            }
            CFRelease(cfstr);
        } else {
            spDebug(10, "xspConvert", "CFStringCreateWithCString failed\n");
        }
    }
#elif defined(MACOS)
    if (TECConvertText != NULL) {
        int ilen;
        ByteCount buf_size;
        ByteCount ail, aol;
        OSStatus status;

        ilen = (int)strlen(input);
        buf_size = MAX(ilen + 1, 32);
        buf = xalloc(buf_size, char);

        while (1) {
            if ((status = TECConvertText(conv->ec, (ConstTextPtr)input, ilen + 1,
                                         &ail, (TextPtr)buf, buf_size, &aol)) == noErr) {
                break;
            }
            if (status != kTECOutputBufferFullStatus) {
                return NULL;
            }
            buf_size += ilen / 4;
            buf = xrealloc(buf, buf_size, char);
        }
    }
#elif defined(_WIN32) && !defined(HAVE_ICONV_H)
    if (streq(conv->icode, conv->ocode)) {
        if (streq(conv->ocode, "UTF-16")) {
            buf = (char *)xspUnicode16Clone((const unsigned short *)input);
        } else {
            buf = strclone(input);
        }
    } else {
        int nchar;
        int buf_size;
        
        spDebug(80, "xspConvert", "icode = %s, ocode = %s\n",
                conv->icode, conv->ocode);
        
        if (streq(conv->icode, "ACP")) {
            nchar = MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, (LPSTR)input,
                                        -1, NULL, 0);

            buf_size = (nchar + 1) * 2;
            buf = xalloc(buf_size, char);

            if (MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, (LPSTR)input, 
                                    -1, (LPWSTR)buf, nchar) == 0) {
                xfree(buf); buf = NULL;
            }

            if (buf != NULL) {
                if (streq(conv->ocode, "UTF-8")) {
                    int obuf_size;
                    char *obuf;
                    
                    if ((obuf_size = spUnicode16ToUTF8((unsigned short *)buf, NULL, 0)) > 0) {
                        spDebug(80, "xspConvert", "obuf_size = %d\n", obuf_size);
                        obuf = xalloc(obuf_size, char);
                        spUnicode16ToUTF8((unsigned short *)buf, (unsigned char *)obuf, obuf_size);

                        xfree(buf);
                        buf = obuf;
                    }
                }
            }
        } else {
            int ibuf_size = 0;
            char *ibuf;
            
            if (streq(conv->icode, "UTF-8")) {
                if ((ibuf_size = spUTF8ToUnicode16((unsigned char *)input, NULL, 0)) <= 0) {
                    return NULL;
                }
                spDebug(80, "xspConvert", "ibuf_size = %d\n", ibuf_size);
                ibuf = xalloc(ibuf_size, char);
                
                spUTF8ToUnicode16((unsigned char *)input,
                                  (unsigned short *)ibuf, ibuf_size);
            } else {
                /* UTF-16 */
                ibuf = (char *)input;
            }
            
            if (streq(conv->ocode, "UTF-16")) {
                buf = (char *)xspUnicode16Clone((const unsigned short *)ibuf);
            } else {
                nchar = WideCharToMultiByte(CP_ACP, WC_COMPOSITECHECK, (LPCWSTR)ibuf,
                                            -1, NULL, 0, NULL, NULL);
            
                buf_size = nchar + 1;
                buf = xalloc(buf_size, char);
            
                if (WideCharToMultiByte(CP_ACP, WC_COMPOSITECHECK, (LPCWSTR)ibuf, 
                                        -1, (LPSTR)buf, nchar, NULL, NULL) == 0) {
                    xfree(buf); buf = NULL;
                }
            }

            if (ibuf_size > 0) {
                xfree(ibuf);
            }
        }
    }
#elif defined(HAVE_ICONV_H)
    if (conv->cd != (iconv_t)(-1)) {
        int ilen;
        int buf_size;
        size_t isize, osize;
        char *iptr, *optr;

        iptr = (char *)input;
        ilen = (int)strlen(iptr);
        isize = ilen;

        buf_size = isize + 1;
        buf = xalloc(buf_size, char);
        
        while (1) {
            optr = buf;
            osize = buf_size - 1;
        
            spDebug(80, "xspConvert", "isize = %d, osize = %d, buf_size = %d\n", isize, osize, buf_size);

            if (iconv(conv->cd, &iptr, &isize, &optr, &osize) == -1) {
                spDebug(80, "xspConvert", "iconv error: errno = %d (E2BIG = %d), isize = %d, osize = %d\n",
                        errno, E2BIG, isize, osize);
                if (errno == E2BIG) {
                    buf_size += (osize + 2);
                    buf = xrealloc(buf, buf_size, char);
                } else {
                    xfree(buf); buf = NULL;
                    break;
                }
            } else {
                *optr = '\0';
                break;
            }
            
            iptr = (char *)input;
            isize = ilen;
        }
    }
#endif

    return buf;
}

void spCloseConverter(spConverter conv)
{
    if (conv == NULL) return;
    
#if defined(SP_MAC_USE_CFSTR_CONVERTER)
    /* do nothing */
#elif defined(MACOS)
    if (TECDisposeConverter != NULL) {
        TECDisposeConverter(conv->ec);
    }
#elif defined(_WIN32) && !defined(HAVE_ICONV_H)
#elif defined(HAVE_ICONV_H)
    iconv_close(conv->cd);
#endif

    xfree(conv);
    
    return;
}

spBool spGetTimeNString(double sec, spTimeFormat time_format, char *buf, int bufsize)
{
    int frac;
    double hour, minute;
    
    if (buf == NULL) return SP_FALSE;

    switch (time_format) {
      case SP_TIME_FORMAT_POINT:
#if defined(HAVE_SPRINTF_S)
        sprintf_s(buf, bufsize, "%ld", (long)spRound(sec));
#elif defined(HAVE_SNPRINTF)
        snprintf(buf, bufsize, "%ld", (long)spRound(sec));
#else
        sprintf(buf, "%ld", (long)spRound(sec));
#endif
        break;
        
      case SP_TIME_FORMAT_SEC:
#if defined(HAVE_SPRINTF_S)
        sprintf_s(buf, bufsize, "%f", sec);
#elif defined(HAVE_SNPRINTF)
        snprintf(buf, bufsize, "%f", sec);
#else
        sprintf(buf, "%f", sec);
#endif
        break;
      case SP_TIME_FORMAT_FLOORED_SEC:
#if defined(HAVE_SPRINTF_S)
        sprintf_s(buf, bufsize, "%.0f", spRound(sec));
#elif defined(HAVE_SNPRINTF)
        snprintf(buf, bufsize, "%.0f", spRound(sec));
#else
        sprintf(buf, "%.0f", spRound(sec));
#endif
        break;

      case SP_TIME_FORMAT_MSEC:
#if defined(HAVE_SPRINTF_S)
        sprintf_s(buf, bufsize, "%.3f", sec * 1000.0);
#elif defined(HAVE_SNPRINTF)
        snprintf(buf, bufsize, "%.3f", sec * 1000.0);
#else
        sprintf(buf, "%.3f", sec * 1000.0);
#endif
        break;
      case SP_TIME_FORMAT_FLOORED_MSEC:
#if defined(HAVE_SPRINTF_S)
        sprintf_s(buf, bufsize, "%.0f", spRound(sec * 1000.0));
#elif defined(HAVE_SNPRINTF)
        snprintf(buf, bufsize, "%.0f", spRound(sec * 1000.0));
#else
        sprintf(buf, "%.0f", spRound(sec * 1000.0));
#endif
        break;

      case SP_TIME_FORMAT_SEPARATED_SEC:
      case SP_TIME_FORMAT_FLOORED_SEPARATED_SEC:
        
        hour = floor(sec / 3600.0);
        sec -= hour * 3600.0;
        minute = floor(sec / 60.0);
        sec -= minute * 60.0;

        if (time_format == SP_TIME_FORMAT_SEPARATED_SEC) {
            frac = (int)spRound(1000000.0 * spFrac(sec));
            if (frac >= 1000000) {
                frac -= 1000000;
                sec++;
            }
            /*spDebug(100, "spGetTimeString", "sec = %f, frac = %d\n", sec, frac);*/
#if defined(HAVE_SPRINTF_S)
            sprintf_s(buf, bufsize, "%.0f:%02d:%02d.%06d", hour, (int)minute, (int)sec, frac);
#elif defined(HAVE_SNPRINTF)
            snprintf(buf, bufsize, "%.0f:%02d:%02d.%06d", hour, (int)minute, (int)sec, frac);
#else
            sprintf(buf, "%.0f:%02d:%02d.%06d", hour, (int)minute, (int)sec, frac);
#endif
        } else {
#if defined(HAVE_SPRINTF_S)
            sprintf_s(buf, bufsize, "%.0f:%02d:%02d", hour, (int)minute, (int)spRound(sec));
#elif defined(HAVE_SNPRINTF)
            snprintf(buf, bufsize, "%.0f:%02d:%02d", hour, (int)minute, (int)spRound(sec));
#else
            sprintf(buf, "%.0f:%02d:%02d", hour, (int)minute, (int)spRound(sec));
#endif
        }
        break;

      default:
        return SP_FALSE;
    }

    return SP_TRUE;
}

spBool spGetTimeString(double sec, spTimeFormat time_format, char *buf)
{
    return spGetTimeNString(sec, time_format, buf, SP_MAX_LINE);
}

spTimeFormat spConvertTimeString(char *buf, spTimeFormat time_format, double *sec)
{
    int nscan;
    int ihour = 0, iminute = 0;
    double value = 0.0;
    double weight = 1.0;
    
    if (strnone(buf) || sec == NULL) return SP_TIME_FORMAT_UNKNOWN;

    if (strchr(buf, ':') != NULL) {
        time_format = SP_TIME_FORMAT_SEPARATED_SEC;
    }
    
    switch (time_format) {
      case SP_TIME_FORMAT_MSEC:
      case SP_TIME_FORMAT_FLOORED_MSEC:
        weight = 1000.0;

      case SP_TIME_FORMAT_POINT:
      case SP_TIME_FORMAT_SEC:
      case SP_TIME_FORMAT_FLOORED_SEC:
#if defined(HAVE_SSCANF_S)
        nscan = sscanf_s(buf, "%lf", &value);
#else
        nscan = sscanf(buf, "%lf", &value);
#endif
        if (nscan < 1) {
            return SP_TIME_FORMAT_UNKNOWN;
        }
        *sec = value / weight;
        break;

      case SP_TIME_FORMAT_SEPARATED_SEC:
      case SP_TIME_FORMAT_FLOORED_SEPARATED_SEC:
#if defined(HAVE_SSCANF_S)
        nscan = sscanf_s(buf, "%d:%d:%lf", &ihour, &iminute, &value);
#else
        nscan = sscanf(buf, "%d:%d:%lf", &ihour, &iminute, &value);
#endif
        if (nscan < 1) {
            return SP_TIME_FORMAT_UNKNOWN;
        } else if (nscan <= 1) {
            ihour = 0; iminute = 0; value = 0.0;
#if defined(HAVE_SSCANF_S)
            sscanf_s(buf, "%lf", &value);
#else
            sscanf(buf, "%lf", &value);
#endif
        } else if (nscan <= 2) {
            ihour = 0; iminute = 0; value = 0.0;
#if defined(HAVE_SSCANF_S)
            sscanf_s(buf, "%d:%lf", &iminute, &value);
#else
            sscanf(buf, "%d:%lf", &iminute, &value);
#endif
        }
        *sec = 3600.0 * (double)ihour + 60.0 * (double)iminute + value;
        /*spDebug(80, "spConvertTimeFormatString", "SEC: buf = %s, sec = %f, value = %f\n", buf, *sec, value);*/
        break;

      default:
        return SP_TIME_FORMAT_UNKNOWN;
    }
    
    return time_format;
}

static void checkTimeFormatStringPrefix(int len, const char *format,
                                        spBool *floored_flag, spBool *separated_flag)
{
    if (len > 0) {
        if (len > 1) {
            if (format[0] == 'f') {
                *floored_flag = SP_TRUE;
            } else if (format[0] == 's') {
                *separated_flag = SP_TRUE;
            }
            if (format[1] == 'f') {
                *floored_flag = SP_TRUE;
            } else if (format[1] == 's') {
                *separated_flag = SP_TRUE;
            }
        } else {
            if (format[0] == 'f') {
                *floored_flag = SP_TRUE;
            } else if (format[0] == 's') {
                *separated_flag = SP_TRUE;
            }
        }
    }

    return;
}

spBool spGetTimeFormatNString(spTimeFormat time_format, char *format, int formatsize)
{
    char floored_string[4];
    char separated_string[4];
    char time_string[8];
    
    if (format == NULL) return SP_FALSE;

    if (time_format & SP_TIME_FORMAT_SEPARATED_MASK) {
        spStrCopy(separated_string, sizeof(separated_string), "s");
    } else {
        spStrCopy(separated_string, sizeof(separated_string), "");
    }
    if (time_format & SP_TIME_FORMAT_FLOORED_MASK) {
        spStrCopy(floored_string, sizeof(floored_string), "f");
    } else {
        spStrCopy(floored_string, sizeof(floored_string), "");
    }

    if (time_format & SP_TIME_FORMAT_POINT_MASK) {
        spStrCopy(time_string, sizeof(time_string), "point");
    } else if (time_format & SP_TIME_FORMAT_MSEC_MASK) {
        spStrCopy(time_string, sizeof(time_string), "msec");
    } else {
        spStrCopy(time_string, sizeof(time_string), "sec");
    }
    
#if defined(HAVE_SPRINTF_S)
    sprintf_s(format, formatsize, "%s%s%s", floored_string, separated_string, time_string);
#elif defined(HAVE_SNPRINTF)
    snprintf(format, formatsize, "%s%s%s", floored_string, separated_string, time_string);
#else
    sprintf(format, "%s%s%s", floored_string, separated_string, time_string);
#endif

    return SP_TRUE;
}

spBool spGetTimeFormatString(spTimeFormat time_format, char *format)
{
    return spGetTimeFormatNString(time_format, format, SP_MAX_LINE);
}

spBool spConvertTimeFormatString(const char *format, spTimeFormat *time_format)
{
    int len;
    spBool msec_flag = SP_FALSE;
    spBool floored_flag = SP_FALSE;
    spBool separated_flag = SP_FALSE;
    
    if (strnone(format) || time_format == NULL) return SP_FALSE;

    len = (int)strlen(format);
    spDebug(50, "spConvertTimeFormatString", "format = %s, len = %d\n", format, len);

    if (len >= 4 && streq(format + (len - 4), "msec")) {
        msec_flag = SP_TRUE;
        checkTimeFormatStringPrefix(len - 4, format, &floored_flag, &separated_flag);
    } else if (len >= 3 && streq(format + (len - 3), "sec")) {
        checkTimeFormatStringPrefix(len - 3, format, &floored_flag, &separated_flag);
    } else if (len >= 1 && streq(format + (len - 1), "m")) {
        msec_flag = SP_TRUE;
        checkTimeFormatStringPrefix(len - 1, format, &floored_flag, &separated_flag);
    } else if (len >= 1 && streq(format + (len - 1), "s")) {
        checkTimeFormatStringPrefix(len - 1, format, &floored_flag, &separated_flag);
    } else if (streq(format, "point")) {
        *time_format = SP_TIME_FORMAT_POINT;
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }

    if (msec_flag == SP_TRUE) {
        *time_format = SP_TIME_FORMAT_MSEC_MASK;
    } else {
        *time_format = SP_TIME_FORMAT_SEC_MASK;
    }
    
    if (floored_flag == SP_TRUE) {
        *time_format |= SP_TIME_FORMAT_FLOORED_MASK;
    }
    if (separated_flag == SP_TRUE) {
        *time_format |= SP_TIME_FORMAT_SEPARATED_MASK;
    }

    return SP_TRUE;
}
